var CC=Components.classes;
var CI=Components.interfaces;

// two helper functions - read line from input stream and remove
// whitespaces at its start
function readString(str){
 var s='';
 if(!str.available())return null;
 while(str.available()>0){
  t=str.read(1);
  if(t=='\n')return s;
  s+=t;
 }
 return s;
}

function trimLine(line){
 while(line.length>0 &&(line[0]==' '||line[0]=='\t'))
  line=line.substr(1,line.length-1);
 return line;
}

//return a value of header field
//case ignored
function getInfo(headers, type){
 return getSubInfo(headers, type, '');
}

// get an argument from string like this 'Type:  subtype=value;subtype2=value'
function getSubInfo(headers, type, subtype){
 const mh=CC['@mozilla.org/network/mime-hdrparam;1'].getService(CI.nsIMIMEHeaderParam);
 type = type.toLowerCase();
 for(i=0;i<headers.length;i++){
  t = headers[i].toLowerCase();
  m=t.search('^'+type+'\s*:');
  if(m==0){
   var ln={value:null};
   var chset='UTF-8';
   l=headers[i].indexOf(':');
   l=headers[i].substr(l+1,headers[i].length-l);
   try{
    t=mh.getParameter(l,subtype,chset,false,ln);
   }catch(e){continue;}
   if(t.length>0){
    //remove quotes, if any
    if(t.length>1 && t[0]=='"' && t[t.length-1]=='"')
     return t.substr(1,t.length-2);
    return t;
   }
  }
 }
 return '';
}

function MessageParser(msg){
 this.ctype = null;
 this.bnd = null;
 this.single = null;
 this.message = null;
 this.hasmore = null;
 this.headers = null;
}

MessageParser.prototype.messageHeaders = function(){
 return this.headers;
}

// internal function to read message headers into array of strings
MessageParser.prototype.getHeaders = function(is){
 var line="";
 var hdr="";
 var headers=[];
 while(is.available()>0){
  line=readString(is);
  if(line[line.length-1]=='\r')
   line=line.substr(0,line.length-1);
  if(line[0]==' '||line[0]=='\t'){
   hdr += ' '+trimLine(line);
  }else if(hdr.length>0){
   headers.push(hdr);
   hdr = line;
  }else if(hdr.length==0){
   hdr = line;
  }
  if(line.length == 0)break;
 }
 if(hdr.length>0)headers.push(hdr);
 return headers;
}

// parse message and store some information for internal use
// MUST BE CALLED BEFORE hasNext OR getMessage*

MessageParser.prototype.parseMessage = function(message){
 var is=CC['@mozilla.org/scriptableinputstream;1'].createInstance(CI.nsIScriptableInputStream);
 is.init(message);
 this.message = is;
 this.headers = this.getHeaders(is);
 this.ctype = getInfo(this.headers, "Content-Type").toLowerCase();
 if(this.ctype.search('^multipart/')==0){
  this.bnd = getSubInfo(this.headers, "Content-Type", "boundary");
  this.single = false;
  var bnd='--'+this.bnd;
  while((l=readString(is))!=null){
   if(l==bnd || l==(bnd+'\r'))break;
  }
 } else {
  this.single = true;
 }
 this.hasmore = true;
}

// return true if there is one or more messages
// or message parts can be read, false otherwise

MessageParser.prototype.hasNext = function(){
 return this.hasmore;
}

// read message as string and stores
// info about it in 'info' array - see getMessage()
// for more details

MessageParser.prototype.getMessageString = function(info){
 var out='';
 if(!this.hasmore)
   return null;
 if(this.single){
   info[0] = '';
   info[1] = this.ctype;
   h = getInfo(this.headers, "Content-Transfer-Encoding");
   for(;;){
    l = readString(this.message);
    if(l == null)break;
    if(h==null||h==''||h=='8bit'||h=='7bit'){
     out += l + '\n';
    }else if(h=='base64'){
     out += DecodeB64LineToString(l);
    }else if(h=='quoted-printable'){
     out += DecodeQPLineToString(l);
    }
   }
   this.hasmore = false;
 }else{
  hdrs = this.getHeaders(this.message);
  info[0] = getInfo(hdrs, 'Content-ID');
  info[1] = getInfo(hdrs, 'Content-Type');
  h = getInfo(hdrs, 'Content-Transfer-Encoding');
  for(;;){
    l = readString(this.message);
    if(l == null){this.hasmore=null;break;}
    if((l == ('--'+this.bnd))||(l == ('--'+this.bnd+'\r'))){
     this.hasmore = true;
     break;
    }
    if((l == ('--'+this.bnd+'--'))||(l == ('--'+this.bnd+'--\r'))){
     this.hasmore = false;
     break;
    }
    if(h==null||h==''||h=='8bit'||h=='7bit'){
     out += l + '\n';
    }else if(h=='base64'){
     out += DecodeB64LineToString(l);
    }else if(h=='quoted-printable'){
     out += DecodeQPLineToString(l);
    }
  }
 }
 return out;
}

// read message as array of codes to data and stores
// info about it in 'info' array - see getMessage()
// for more details

MessageParser.prototype.getMessageArray = function(info,data){
 var out='';
 if(!this.hasmore)
   return null;
 if(this.single){
   info[0] = '';
   info[1] = this.ctype;
   h = getInfo(this.headers, "Content-Transfer-Encoding");
   for(;;){
    l = readString(this.message);
    if(l == null)break;
    if(h==null||h==''||h=='8bit'||h=='7bit'){
     for(i=0;i<l.length;i++)data.push(l.charCodeAt(i));
     data.push(10);
    }else if(h=='base64'){
     DecodeB64Line(l,data);
    }else if(h=='quoted-printable'){
     DecodeQPLine(l,data);
    }
   }
   this.hasmore = false;
 }else{
  hdrs = this.getHeaders(this.message);
  info[0] = getInfo(hdrs, 'Content-ID');
  info[1] = getInfo(hdrs, 'Content-Type');
  h = getInfo(hdrs, 'Content-Transfer-Encoding');
  for(;;){
    l = readString(this.message);
    if(l == null){this.hasmore=null;break;}
    if((l == ('--'+this.bnd))||(l == ('--'+this.bnd+'\r'))){
     this.hasmore = true;
     break;
    }
    if((l == ('--'+this.bnd+'--'))||(l == ('--'+this.bnd+'--\r'))){
     this.hasmore = false;
     break;
    }
    if(h==null||h==''||h=='8bit'||h=='7bit'){
    }else if(h=='base64'){
     DecodeB64Line(l,data);
    }else if(h=='quoted-printable'){
     DecodeQPLine(l,data);
    }
  }
 }
}

// getMessage - returns data for message and stores
// some info about it in 'info' array:
// info[0] - Content-ID of the message
// info[1] - Content-Type of the message
// info[2] - message is passed as string or not,
//           true -  return result is string,
//           false - return result is array of codes.

MessageParser.prototype.getMessage = function(info){
 var out='';
 var out2=[];
 if(!this.hasmore)
   return null;
 if(this.single){
   info[0] = '';
   info[1] = this.ctype;
   if(info[1].search("^text/")==0)
    info[2]=true;
   else
    info[2]=false;
   h = getInfo(this.headers, "Content-Transfer-Encoding");
   for(;;){
    l = readString(this.message);
    if(l == null)break;
    if(h==null||h==''||h=='8bit'||h=='7bit'){
     if(info[2]){
      out += l + '\n';
     }else{
      for(i=0;i<l.length;i++)out2.push(l.charCodeAt(i));
      out2.push(10);
     }
    }else if(h=='base64'){
     if(info[2])
      out += DecodeB64LineToString(l);
     else
      DecodeB64Line(l, out2);
    }else if(h=='quoted-printable'){
     if(info[2])
      out += DecodeQPLineToString(l);
     else
      DecodeQPLine(l,out2);
    }
   }
   this.hasmore = false;
 }else{
  hdrs = this.getHeaders(this.message);
  info[0] = getInfo(hdrs, 'Content-ID');
  info[1] = getInfo(hdrs, 'Content-Type');
  if(info[1].search("^text/")==0)
   info[2]=true;
  else
   info[2]=false;
  h = getInfo(hdrs, 'Content-Transfer-Encoding');
  for(;;){
    l = readString(this.message);
    if(l == null){this.hasmore=null;break;}
    if((l == ('--'+this.bnd))||(l == ('--'+this.bnd+'\r'))){
     this.hasmore = true;
     break;
    }
    if((l == ('--'+this.bnd+'--'))||(l == ('--'+this.bnd+'--\r'))){
     this.hasmore = false;
     break;
    }
    if(h==null||h==''||h=='8bit'||h=='7bit'){
     if(info[2]){
      out += l + '\n';
     }else{
      for(i=0;i<l.length;i++)out2.push(l.charCodeAt(i));
      out2.push(10);
     }
    }else if(h=='base64'){
     if(info[2])
      out += DecodeB64LineToString(l);
     else
      DecodeB64Line(l, out2);
    }else if(h=='quoted-printable'){
     if(info[2])
      out += DecodeQPLineToString(l);
     else
      DecodeQPLine(l,out2);
    }
  }
 }
 if(info[2])
  return out;
 else
  return out2;
}
